from inspect import CO_GENERATOR
import numpy as np
from matplotlib import patches, text, lines
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
import pandas as pd


class Visualizer(object):
    """The visualizer object stores parameters and methods for generating an
    animation of a particular run.

    Parameters
    ----------
    data : pd.DataFrame
        Dataframe containing simulation data
    pend : pendsim.sim.Pendulum
        Pendulum object to visualize. The animation scales with the length and masses
        of the pendulum object. To generate an accurate visualization, just pass the same
        pendulum that was simulated.
    speed : int, optional
        The number of frames to "skip" in the animation. Real-time animation is rather
        difficult with matplotlib, because the draw speed of the animator varies between
        systems. So if your simulation timestep is very small, the animation can run quite
        slowly.      by default 2
    """

    def __init__(self, data: pd.DataFrame, pend, speed: int = 2) -> None:

        self.data, self.pend = data, pend
        self.speed = speed
        self.cart_w = np.sqrt(2 * self.pend.M) * 0.5
        self.cart_h = np.sqrt(1 / (2) * self.pend.M) * 0.5
        # Pendulum Params
        self.p_rad = np.sqrt(self.pend.m) * 0.5 / 3
        # Display Params
        self.xmax = self.data.loc(axis=1)["state", "x"].values.max() * 1.1
        self.xmin = self.data.loc(axis=1)["state", "x"].values.min() * 1.1
        self.ymax = (self.pend.l + self.cart_h) * 1.3
        self.ymin = -self.pend.l * 1.3
        self.f_width = 0.02

    def _draw_cart(
        self,
        cart: patches.Rectangle,
        mass: patches.Circle,
        line: lines.Line2D,
        xi: float,
        thetai: float,
    ):
        """
        Draw cart on `cart`, `mass`, and `line` objects with cart position xi and
        pendulum position thetai.
        """
        # adjust height when rendering
        cart.set_xy((xi - self.cart_w * 0.5, self.cart_h))
        # mass xy
        massx = xi - self.pend.l * np.sin(thetai)
        massy = self.cart_h + self.pend.l * np.cos(thetai)
        mass.set_center((massx, massy))
        # line
        linexy = np.array([[massx, xi], [massy, self.cart_h]])
        line.set_data(linexy)

    def _draw_objs(self, lc="black", ls="-"):
        """Create objects to draw

        Parameters
        ----------
        lc : str, optional
            matplotlib Line color, by default "black"
        ls : str, optional
            matplotlib Line style, by default "-"

        Returns
        -------
        Tuple[patches.Rectangle, patches.Circle, patches.Line2D]
            Tuple of the cart, pendulum, and connecting line objects
        """
        cart = patches.Rectangle(
            (-self.cart_w * 0.5, self.cart_h),
            width=self.cart_w,
            height=-self.cart_h,
            fc="white",
            ec=lc,
            ls=ls,
        )
        # The pendulum mass
        mass = patches.Circle((0, 0), radius=self.p_rad, fc="white", ec=lc, ls=ls)
        # The line connecting cart to pend mass
        line = lines.Line2D(
            [0, 1],
            [0, 1],
            c="black",
            linestyle="-",
        )
        return cart, mass, line

    def draw_force(self, obj, u, cart_x, ydist):
        """Draw the external force on the cart.

        Parameters
        ----------
        obj : matplotlib FancyArrow object
            The force line object
        u : float
            magnitude of force to represent
        cart_x : float
            cart position
        ydist : float
            distance vertically "up" the cart to draw the force. i.e. 0 draws
            the force at the very bottom of the cart along the ground, 1.0 draws
            the force at the very top of the cart
        """
        if u > 0.0:
            beg = (cart_x - 0.5 * self.cart_w, ydist * self.cart_h)
            end = (
                cart_x - 0.5 * self.cart_w - np.sqrt(0.1 * np.abs(u)),
                ydist * self.cart_h,
            )
            obj.set_xy((beg, end))
            obj.set_linewidth(np.sqrt(np.abs(u)))
            obj.set_visible(True)
        elif u < 0.0:
            beg = (cart_x + 0.5 * self.cart_w, ydist * self.cart_h)
            end = (
                cart_x + 0.5 * self.cart_w + np.sqrt(0.1 * np.abs(u)),
                ydist * self.cart_h,
            )
            obj.set_xy((beg, end))
            obj.set_linewidth(np.sqrt(np.abs(u)))
            obj.set_visible(True)
        else:
            obj.set_xy(((0, 0), (1, 1)))
            obj.set_linewidth(0)
            obj.set_visible(False)

    def animate(
        self,
        pltdata={},
        interval=10,
        draw_fbd=False,
        data_stretch=False,
        figsize=(8, 4.5),
    ):
        if pltdata:
            fig, ax = plt.subplots(nrows=2, figsize=figsize)
            ax0, ax1 = ax[0], ax[1]
        else:
            fig, ax0 = plt.subplots(figsize=figsize)

        if data_stretch:
            yls, yus = [], []
            for reskey in pltdata.keys():
                yls.append(self.data[reskey].values.min())
                yus.append(self.data[reskey].values.max())
            yl = min(yls)
            yu = max(yus)
            xl = self.data[reskey].index.min()
            xu = self.data[reskey].index.max()

            ax1.set_ylim((yl, yu))
            ax1.set_xlim((xl, xu))

        # axis setup
        ax0.set_xlim(self.xmin - self.pend.l * 2, self.xmax + self.pend.l * 2)
        ax0.set_ylim(self.ymin, self.ymax)
        ax0.set_aspect("equal")
        n_frames = (
            np.floor(len(self.data.index.values.tolist()) / self.speed).astype(int) - 10
        )
        # Initialize objects
        cart, mass, line = self._draw_objs()
        # Line for external force
        ext_force = patches.FancyArrow(0, 0, 1, 1, ec="red")
        # Line for control force
        ctrl_force = patches.FancyArrow(0, 0, 1, 1, ec="blue")

        if draw_fbd:
            pRx_f = patches.FancyArrow(0, 0, 1, 1, ec="k", zorder=4)
            pRy_f = patches.FancyArrow(0, 0, 1, 1, ec="k", zorder=4)
            pG_f = patches.FancyArrow(0, 0, 1, 1, ec="k", zorder=4)
            cRx_f = patches.FancyArrow(0, 0, 1, 1, ec="k", zorder=4)
            cRy_f = patches.FancyArrow(0, 0, 1, 1, ec="k", zorder=4)
            cG_f = patches.FancyArrow(0, 0, 1, 1, ec="k", zorder=4)
            cN_f = patches.FancyArrow(0, 0, 1, 1, ec="k", zorder=4)
            fbd_draws = (pRx_f, pRy_f, pG_f, cRx_f, cRy_f, cG_f, cN_f)

        ground = patches.Rectangle((-1000, -2000), 2000, 2000, fc="grey")
        # ground
        ground.set_zorder(-1)
        # Time text
        time_text = text.Annotation("", (4, 28), xycoords="axes points")

        plots = []
        for name, attrs in pltdata.items():
            if attrs["type"] == "line":
                (plot,) = ax1.plot(
                    [],
                    [],
                    label=attrs["label"],
                    linestyle=attrs["linestyle"],
                    color=attrs["color"],
                )
                plots.append(plot)
            elif attrs["type"] == "scatter":
                plot = ax1.scatter(
                    [], [], label=attrs["label"], c=attrs["color"], edgecolors=None
                )
                plots.append(plot)
            else:
                raise ValueError("Wrong type or no type given.")

        def _init():
            plist = []
            ax0.add_patch(cart)
            ax0.add_patch(mass)
            ax0.add_artist(line)
            ax0.add_patch(ext_force)
            ax0.add_patch(ctrl_force)
            ax0.add_patch(ground)
            ax0.add_artist(time_text)
            if draw_fbd:
                for fbd in fbd_draws:
                    ax0.add_patch(fbd)
                plist.extend(fbd_draws)
            plist = [ground, cart, mass, line, ext_force, ctrl_force, time_text]
            plist.extend(plots)
            return plist

        def _animate(i):
            i = np.floor(i * self.speed).astype(int)

            retobjs = []
            # limits for y-axis

            if pltdata:
                scyall = [0]
                for (name, attrs), sc in zip(pltdata.items(), plots):
                    l = max(i - attrs["plotpoints"], 0)
                    scx = self.data.index[l:i]
                    scy = self.data[name].values[l:i]
                    if attrs["type"] == "scatter":
                        sc.set_offsets(np.column_stack([scx, scy]))
                    elif attrs["type"] == "line":
                        sc.set_data(scx, scy)
                    scyall.extend(list(scy))
                retobjs.extend(plots)
                yl = min(-0.1, min(scyall))
                yu = max(0.1, max(scyall))
                xl = self.data.index[
                    max(i - max([p["plotpoints"] for p in pltdata.values()]), 0)
                ]
                xu = self.data.index[i] + 1e-5
                if not data_stretch:
                    ax1.set_ylim((yl, yu))
                    ax1.set_xlim((xl, xu))
                ax1.legend(loc=2)

            # draw cart
            state_xi = list(self.data[("state", "x")].values)[i]
            state_ti = list(self.data[("state", "t")].values)[i]
            self._draw_cart(cart, mass, line, state_xi, state_ti)
            # external force
            self.draw_force(
                ext_force,
                list(self.data[("forces", "forces")].values)[i],
                state_xi,
                0.6,
            )
            self.draw_force(
                ctrl_force,
                list(self.data[("control action", "control action")].values)[i],
                state_xi,
                0.5,
            )
            time_text.set_text(r"t=" + str(round(self.data.index[i], 3)))

            # fbds
            if draw_fbd:
                # pend x
                px = state_xi - self.pend.l * np.sin(state_ti)
                # pend y
                py = self.cart_h + self.pend.l * np.cos(state_ti)
                # draw reaction force (pendulum)
                self.draw_pend_fbd(
                    pRx_f,
                    self.data[("forces", "pRx")].values[i],
                    np.array([1, 0]),
                    np.array([px, py]),
                )
                self.draw_pend_fbd(
                    pRy_f,
                    self.data[("forces", "pRy")].values[i],
                    np.array([0, 1]),
                    np.array([px, py]),
                )
                self.draw_pend_fbd(
                    pG_f,
                    self.data[("forces", "pG")].values[i],
                    np.array([0, 1]),
                    np.array([px, py]),
                )
                retobjs.extend((pRx_f, pRy_f, pG_f))
                # draw reaction force (cart)
                cx, cy = state_xi, self.cart_h
                self.draw_cart_fbd(
                    cRx_f,
                    self.data[("forces", "cRx")].values[i],
                    np.array([1, 0]),
                    np.array([cx, cy]),
                )
                self.draw_cart_fbd(
                    cRy_f,
                    self.data[("forces", "cRy")].values[i],
                    np.array([0, 1]),
                    np.array([cx, cy]),
                )
                self.draw_cart_fbd(
                    cG_f,
                    self.data[("forces", "cG")].values[i],
                    np.array([0, 1]),
                    np.array([cx, cy]),
                )
                self.draw_cart_fbd(
                    cN_f,
                    self.data[("forces", "cN")].values[i],
                    np.array([0, 1]),
                    np.array([cx, cy]),
                )

                retobjs.extend((cRx_f, cRy_f, cG_f, cN_f))

            retobjs.extend([ground, cart, mass, line, ext_force, ctrl_force, time_text])
            return retobjs

        return FuncAnimation(
            fig,
            _animate,
            frames=n_frames,
            init_func=_init,
            blit=True,
            interval=interval,
        )

    def draw_cart_fbd(self, obj, f, direc, pos):
        direc = sign(f) * direc
        # initial position plus offset for cart radius
        offsetbeg = pos
        # set beginning/end
        beg, end = offsetbeg, offsetbeg + np.sqrt(self.f_width * np.abs(f)) * direc
        if np.abs(f) > 0:
            obj.set_xy((beg, end))
            obj.set_linewidth(np.sqrt(np.abs(f)))
            obj.set_visible(True)
        else:
            obj.set_visible(False)

    def draw_pend_fbd(self, obj, f, direc, pos):
        direc = sign(f) * direc
        # its initial position plus offset for pend radius
        offsetbeg = pos + direc * self.p_rad
        # set beginning / end
        beg, end = offsetbeg, offsetbeg + np.sqrt(self.f_width * np.abs(f)) * direc
        if np.abs(f) > 0:
            obj.set_xy((beg, end))
            obj.set_linewidth(np.sqrt(np.abs(f)))
            obj.set_visible(True)
        else:
            obj.set_visible(False)


def sign(x):
    if x >= 0:
        return 1
    else:
        return -1



from matplotlib import get_backend
import pendsim, pytest
import numpy as np


@pytest.fixture
def pend():
    return pendsim.sim.Pendulum(3.0, 1.0, 2.0)


@pytest.fixture
def BangBang():
    return pendsim.controller.BangBang(0, 10)


@pytest.fixture
def PID():
    return pendsim.controller.PID(1, 1, 1)


@pytest.fixture
def LQR(pend):
    Q = np.diag([1, 1, 1, 1])
    R = np.atleast_2d([1])
    return pendsim.controller.LQR(pend, 0.01, 10, Q, R)


@pytest.fixture
def NoController():
    return pendsim.controller.NoController()


@pytest.fixture
def policy_args():
    return (np.array([0.1, 0.0, 2.11, 0.02]), 0.02)


def test_policy_BangBang(BangBang, policy_args):
    policy_output = BangBang.policy(*policy_args)
    assert type(policy_output) == tuple
    assert type(policy_output[0]) == float or type(policy_output[0]) == np.float64
    assert type(policy_output[1]) == dict


def test_policy_PID(PID, policy_args):
    policy_output = PID.policy(*policy_args)
    assert type(policy_output) == tuple
    assert type(policy_output[0]) == float or type(policy_output[0]) == np.float64
    assert type(policy_output[1]) == dict


def test_policy_LQR(LQR, policy_args):
    policy_output = LQR.policy(*policy_args)
    assert type(policy_output) == tuple
    assert type(policy_output[0]) == float or type(policy_output[0]) == np.float64
    assert type(policy_output[1]) == dict


def test_policy_NoController(NoController, policy_args):
    policy_output = NoController.policy(*policy_args)
    assert type(policy_output) == tuple
    assert type(policy_output[0]) == float or type(policy_output[0]) == np.float64
    assert type(policy_output[1]) == dict



from typing import Hashable
import numpy as np
from matplotlib import pyplot as plt
from typing import Iterable

def array_to_kv(level1_key: Hashable, level2_keys: Iterable[Hashable], array: np.ndarray) -> dict:
    data={}
    if array.shape[0] != len(level2_keys):
        raise(ValueError('Level 2 keys are not same length as array: {} vs {}'.format(level2_keys, array.shape[0])))
    for n, name in enumerate(level2_keys):
        key = (level1_key, name)
        val = array[n]
        data[key] = val
    return data

def sign(x: float) -> float:
    if x >= 0: return 1.0
    else: return -1.0

def impulse_func(a:float=0.5, t_offset:float=4.0) -> callable:
    return lambda t: 1/(a*np.sqrt(np.pi)) * np.exp(-((t-t_offset)/a)**2)

def sine_func(period=1) -> callable:
    return lambda t: np.sin(period*t)

def plot_input_func(ifunc:callable, t_final:float, ax: plt.Axes) -> None:
    x = np.linspace(0, t_final, 1000)
    y = ifunc(x)
    ax.plot(x, y)


import numpy as np
from scipy import integrate
from collections import defaultdict
from tqdm import tqdm
from multiprocessing import Pool
import pandas as pd
from typing import Tuple
from pendsim.utils import array_to_kv


##### Utility Functions
class Pendulum(object):
    """Class for a pendulum object

    Parameters
    ----------
    M : float
        Cart weight, in kg
    m : float
        Pendulum weight, in kg
    l : float
        Length of pendulum, in m
    g : float, optional
        Gravitational constant, in m/s^2, by default 9.81
    cfric : float, optional
        Cart viscous friction coefficient, by default 0.1
    pfric : float, optional
        Pendulum viscous friction coefficient, by default 0.05
    initial_state : np.ndarray, optional
        Initial pendulum state, by default np.array([0, 0, 0, 0])

    Attributes
    ----------
    jacA : np.array
        The jacobian of the state transition matrix A

    jacB : np.array
        The jacobian of the control matrix B
    """

    def __init__(
        self,
        M: float,
        m: float,
        l: float,
        g: float = 9.81,
        cfric: float = 0.1,
        pfric: float = 0.05,
        initial_state: np.ndarray = np.array([0, 0, 0, 0]),
    ) -> None:

        self.M = M
        self.m = m
        self.l = l
        self.g = g
        self.cfric = cfric
        self.pfric = pfric
        self.y_0 = initial_state
        # jacobians of the system
        # linearized about the upward position x=[0,0,0,0]
        self.jacA = np.array(
            [
                [0, 1.0, 0, 0],
                [0, 0, -self.m * self.g / self.M, 0.0],
                [0, 0, 0, 1.0],
                [0, 0, self.g * (self.m * self.M) / (self.l * self.M), 0.0],
            ]
        )
        self.jacB = np.array([[0.0], [1.0 / self.M], [0.0], [-1.0 / (self.l * self.M)]])

    def system_dynamics(self, t: float, state: np.ndarray) -> np.ndarray:
        """Solve the system dynamics in the form xdot = A\\x

        Because this function is called by `solve_ivp`, we package the control
        input `u` with the state; i.e. the state vector is [x, xdot, theta, thetadot, u]

        input `t` is unused, because the pendulum is modeled as an LTI system.

        Parameters
        ----------
        t : float
            time
        state : np.ndarray
            state, including cart force `u`.

        Returns
        -------
        np.ndarray
            x dot
        """
        # unpack variables
        xd, theta, td, u = state[1], state[2], state[3], state[4]
        m, M, l, g, cfric, pfric = (
            self.m,
            self.M,
            self.l,
            self.g,
            self.cfric,
            self.pfric,
        )
        cost, sint = np.cos(theta), np.sin(theta)
        # solve for derivatives
        xdd = (g * m * sint * cost + u - m * l * td * td * sint) / (
            M + m - m * cost * cost
        )
        tdd = (g * sint + xdd * cost) / l
        # frictions
        xdd += -cfric * xd
        tdd += -pfric * td
        return np.array([xd, xdd, td, tdd, u])

    def solve(
        self, dt: float, system_state: np.ndarray, u: float, solve_args: dict = {}
    ) -> Tuple[np.ndarray, float]:
        """Given a system state, solve the ivp over interval `d

        Parameters
        ----------
        dt : float
            Timestep over which to solve ivp
        system_state : np.ndarray
            the 4-tuple system state [x, xdot, theta, thetadot]
        u : float
            external force applied to base
        solve_args : dict, optional
            arguments supplied to the solver, by default {}

        Returns
        -------
        Tuple[np.ndarray, float]
            tuple of (4-tuple system state, external force)
        """
        # roll state and u into a single array
        y = np.empty((5))
        y[:4] = system_state
        y[4] = u
        # solve system
        sol = integrate.solve_ivp(self.system_dynamics, (0, dt), y, **solve_args)
        return sol.y[:4, -1], sol.y[4, -1]

    def get_energy(self, state: np.ndarray) -> Tuple[float, float, float]:
        """Get energy of a state.

        Parameters
        ----------
        state : np.ndarray
            system state

        Returns
        -------
        Tuple[float, float, float]
            tuple of (kinetic, potential, total) energies.
        """
        xd, t, td = state[1], state[2], state[3]
        ke = (
            0.5 * (self.m + self.M) * xd * xd
            - self.m * self.l * xd * td * np.cos(t)
            + 0.5 * self.m * self.l * self.l * td * td
        )
        # potential energy
        pe = self.m * self.g * self.l * np.cos(t)
        return ke, pe, ke + pe

    def calculate_reaction_forces(
        self, state: np.ndarray, xdd: float, tdd: float
    ) -> tuple:
        """Get reaction forces on each body (cart and pendulum). This can be
        used to construct or visualize the forces that are applied to each body

        Parameters
        ----------
        state : np.ndarray
            system state
        xdd : float
            cart acceleration
        tdd : float
            angular acceleration

        Returns
        -------
        tuple
            (pend reaction force x-dir, pend reaction force y-dir, pend G force,
            cart reaction force x-dir, cart reaction force y-dir, cart G force,
            cart normal force)
        """
        x, xd, theta, td = state[0], state[1], state[2], state[3]
        sint, cost = np.sin(theta), np.cos(theta)
        pRx = self.m * (xdd + self.l * td * td * sint - self.l * tdd * cost)
        pRy = self.m * (-self.l * td * td * cost - self.l * tdd * sint)
        pG = self.m * self.g
        cRy = -pRy
        cRx = -pRx
        cG = self.M * self.g
        cN = -cG - cRy
        return pRx, pRy, pG, cRy, cRx, cG, cN


class Simulation(object):
    """Simulation object

    Parameters
    ----------
    dt : float
        simulation timestep in seconds
    t_final : float
        final simulation time in seconds (simulation happens over the interval [0, t_final])
    force : callable
        a time-dependent forcing function. Must be in the form `f(t) = u` where t is the
        supplied time and u is the corresponding force applied to the cart.
    noise_scale : np.ndarray, optional
        scale of gaussian noise to add to the system state. Is a 4-tuple where each value corresponds
        to the variance of the noise added to each state value, by default None
    """

    def __init__(
        self, dt: float, t_final: float, force: callable, noise_scale: np.ndarray = None
    ):
        self.dt = dt  # time step
        self.t_final = t_final  # end at or before this time
        self.force = force  # forcing function
        self.noise_scale = noise_scale

    def simulate(self, pendulum: Pendulum, controller) -> pd.DataFrame:
        """Simulate the system with the parameters stored in the `Simulation` object.

        User supplies a pendulum and a controller, so that the same simulation conditions
        (timestep, noise, external force, etc) can be tested against multiple pendulums and
        controllers.

        Parameters
        ----------
        pendulum : Pendulum
            Pendulum object under simulation
        controller : [type]
            Controller to run over the simulation

        Returns
        -------
        pd.DataFrame
            The results of the simulation. Calling e.g. results.columns will show all fields
            recorded during the simulation. This is a 2-level Multi-Index dataframe.
        """
        # initial state
        state = pendulum.y_0
        # initialize data
        datas = defaultdict(list)
        statelabels = ["x", "xd", "t", "td"]
        # time
        times = np.arange(start=0, stop=self.t_final, step=self.dt)

        for k, t in tqdm(enumerate(times), total=len(times)):
            data = {}
            force = self.force(t)

            # Control policy
            data.update(array_to_kv("state", statelabels, state))
            if self.noise_scale is not None:
                noisy_state = state + np.random.normal(0, scale=self.noise_scale)
                data.update(array_to_kv("measured state", statelabels, noisy_state))
                data.update(
                    array_to_kv("noise variance", statelabels, self.noise_scale)
                )
                action, controller_data = controller.policy(noisy_state, self.dt)
            else:
                action, controller_data = controller.policy(state, self.dt)
            data.update(controller_data)

            # Simulation Data
            (
                data[("energy", "kinetic")],
                data[("energy", "potential")],
                data["energy", "total"],
            ) = pendulum.get_energy(state)
            data[("forces", "forces")] = force
            data[("control action", "control action")] = action

            # Add action to force
            force += action
            # Calculate Reaction Forces
            deriv_input = np.empty((5,))
            deriv_input[
                0:4,
            ] = state
            deriv_input[
                0,
            ] = action
            deriv = pendulum.system_dynamics(0, deriv_input)
            xdd, tdd = deriv[2], deriv[4]
            pRx, pRy, pG, cRy, cRx, cG, cN = pendulum.calculate_reaction_forces(
                state, xdd, tdd
            )
            data[("state", "xdd")] = xdd
            data[("state", "tdd")] = tdd
            data[("forces", "pRx")] = pRx
            data[("forces", "pRy")] = pRy
            data[("forces", "pG")] = pG
            data[("forces", "cRy")] = cRy
            data[("forces", "cRx")] = cRx
            data[("forces", "cG")] = cG
            data[("forces", "cN")] = cN

            state, _ = pendulum.solve(self.dt, state, force)
            for k, v in data.items():
                datas[k].append(v)

        return pd.DataFrame(datas, index=times)

    def simulate_multiple(self, pendulums, controllers, parallel=True):
        """Simulate pendulums and controllers in a list. This method sees significant speedup
        in parallelization.

        Parameters
        ----------
        pendulums : list
            List of pendulums to simulate
        controllers : list
            List of controllers to simulate
        parallel : bool, optional
            Whether to simulate in parallel across multiple cores, by default True

        Returns
        -------
        pd.DataFrame
            Dataframe of results. This dataframe basically stacks each run in a multi-index,
            where the level0 indices are the multiple runs of the simulation and the level1 indices
            are the timesteps of each run.

        Raises
        ------
        ValueError
            pendulums and controllers must be 1:1, i.e. each pendulum must be supplied with a
            controller, even if controllers or pendulums used across runs are the same.
        """
        if len(pendulums) != len(controllers):
            raise ValueError(
                "pendulums and controllers must have same length. len(pendulums)={}, len(controllers)={}".format(
                    len(pendulums), len(controllers)
                )
            )
        pc = list(zip(pendulums, controllers))
        if parallel:
            results = Pool().map(self.runsim, pc)
            return pd.concat(results, axis=0, keys=list(range(len(results))))
        else:
            allresults = []
            for pendulum, controller in pc:
                results = self.simulate(pendulum, controller)
                allresults.append(results)
            return pd.concat(allresults, axis=0, keys=list(range(len(results))))

    def runsim(self, pend_cont_tuple):
        return self.simulate(*pend_cont_tuple)



import pendsim.sim, pendsim.utils
import copy

import numpy as np
from scipy.signal import cont2discrete

# needed for LQR + UKF
from filterpy.kalman import UnscentedKalmanFilter, sigma_points

# needed for typeannotations
from typing import Tuple

LABELS = ["x", "xd", "t", "td"]


class Controller(object):
    """
    Base class for controllers. A controller executes a `policy` during
    the simulation loop. It takes some measured `state` and takes some
    `action` on that state. The `action` is a force applied to the base
    of the cart.

    A controller's `policy` can include a state estimator or functions
    we can use to record data.

    Methods
    -------
    policy:
        control policy

    do_swingup:
        get action from Astrom's "swing-up" policy

    do_lqr:
        get action from an LQR policy

    get_linear_sys:
        get the linearized system from jacobians

    store_4tuple:
        store a 4-tuple into multi-index

    wrapPi:
        wrap an angular value to [-pi, pi] interval

    create_ukf:
        create an unscented kalman filter for the system

    do_pid:
        get action from a PID control policy

    get_and_store_priors:
        utility function for storing state priors in a moving backwards horizon

    """

    def __init__(self, init_state: np.ndarray):
        """init function for Controller object. Rare that this would be used
        directly.

        Parameters
        ----------
        init_state : np.ndarray
            Initial state.
        """
        self.init_state = init_state
        self.prev_err = 0

    def policy(self, state: np.ndarray, dt: float) -> Tuple[float, dict]:
        raise NotImplementedError

    def do_swingup(
        self, pend: pendsim.sim.Pendulum, state: np.ndarray, k: float
    ) -> float:
        """Implement a swing-up by Energy control method described by Astrom
        and Furuta. Typically, we want to implement the swing-up strategy if
        the pendulum is below some threshold (i.e. theta > pi/2).

        See

        Åström, Karl Johan, and Katsuhisa Furuta. "Swinging up a pendulum by
        energy control." Automatica 36.2 (2000): 287-295.
        at https://doi.org/10.1016/S1474-6670(17)57951-3

        Parameters
        ----------
        pend : pendsim.sim.Pendulum
            The pendulum object containing `m`,`g`,`l` parameters we use to
            estimate the energy of the system.
        state : np.ndarray
            The current state
        k : float
            Swing-up gain

        Returns
        -------
        float
            controller action.
        """
        m, g, l = pend.m, pend.g, pend.l
        E_norm = 2 * m * g * l
        E = m * g * l * (np.cos(state[2]) - 1)  # 0 = upright
        beta = E / E_norm
        u = k * beta * pendsim.utils.sign(state[3] * np.cos(state[2]))
        return -u

    def do_lqr(
        self,
        w: int,
        A: np.ndarray,
        B: np.ndarray,
        Q: np.ndarray,
        R: np.ndarray,
        x: np.ndarray,
    ) -> np.ndarray:
        """Finite-horizon discrete Linear Quadratic Regulator policy.
        An LQR controller produces an optimal control policy over a finite horizon
        according to a quadratic cost over the system state.

        See https://underactuated.mit.edu/lqr.html


        Parameters
        ----------
        w : int
            window over which to perform LQR control
        A : np.ndarray
            linear plant transition matrix
        B : np.ndarray
            linear control matrix
        Q : np.ndarray
            controller gain matrix
        R : np.ndarray
            control action penalty
        x : np.ndarray
            system state

        Returns
        -------
        np.ndarray
            sequence of control actions
        """

        P = [None] * (w + 1)
        P[w] = Q
        for k in range(w, 0, -1):
            p1 = A.T @ P[k] @ A  # (4,4)
            p2 = A.T @ P[k] @ B  # (4,1)
            p3 = R + B.T @ P[k] @ B  # (4,1)
            p3 = np.linalg.pinv(R + B.T @ P[k] @ B)
            p4 = B.T @ P[k] @ A
            P[k - 1] = p1 - p2 @ p3 @ p4 + Q

        u = [None] * w
        for k in range(w):
            c1 = np.linalg.inv(R + B.T @ P[k] @ B)
            c2 = B.T @ P[k] @ A
            u[k] = c1 @ c2 @ x
        return np.squeeze(u)

    def get_linear_sys(
        self, Adot: np.ndarray, Bdot: np.ndarray, dt: float
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Get linearize from jacobians of system with timestep `dt`.


        Parameters
        ----------
        Adot : np.ndarray
            The (4,4) jacobian of the plant matrix A
        Bdot : np.ndarray
            The (4,1) jacobian of the control matrix B
        dt : float
            control timestep

        Returns
        -------
        Tuple[np.ndarray, np.ndarray]
            linearized system matrices A, B.
        """
        A, B = Adot, Bdot
        C, D = np.zeros((1, A.shape[0])), np.zeros((1, 1))
        sys_disc = cont2discrete((A, B, C, D), dt, method="zoh")
        return sys_disc[0], sys_disc[1]

    def store_4tuple(self, level1_key: str, val: np.ndarray) -> dict:
        """Helper function for storing a 4-tuple of values. A level 0 key
        is given, and the level 1 keys are populated with 'x', 'xd', 't','td'
        values. Use this when you need to easily return data from a 4-tuple
        produced by the simulation that maps to each of the 4 state values.

        Example

        >>> val = [1,2,3,4]
        >>> key = "count"
        >>> store4tuple(key, val)
        {
            ('count','x') : 1,
            ('count','xd') : 2,
            ('count','t') : 3,
            ('count','td') : 4,
        }

        Parameters
        ----------
        level1_key : str
            outer key
        val : np.ndarray
            values. Must be shape (4,1) or (4,)

        Returns
        -------
        dict
            dict containing new data values
        """
        labels = ["x", "xd", "t", "td"]
        return pendsim.utils.array_to_kv(level1_key, labels, np.squeeze(val))

    def wrapPi(self, val: float) -> float:
        """Wrap an angle to the interval between [-pi, pi].

        Parameters
        ----------
        val : float
            angle

        Returns
        -------
        float
            wrapped angle
        """
        return (val + np.pi) % (2 * np.pi) - np.pi

    def create_ukf(
        self, dt: float, hx: callable, fx: callable
    ) -> UnscentedKalmanFilter:
        """Create an unscented Kalman filter with state transition functions
        `fx`, measurement function `hx` with timestep between measurements
        estimated as `dt`.

        Parameters
        ----------
        dt : float
            control timestep
        hx : callable
            function describing mapping between sensor inputs and measurements
        fx : callable
            function describing system. Can be non-linear.

        Returns
        -------
        UnscentedKalmanFilter
            UKF object with sensible defaults for sigma points.
        """
        n = 4
        points = sigma_points.MerweScaledSigmaPoints(
            n,
            alpha=1e-4,
            beta=2,
            kappa=3 - n,
        )

        kf = UnscentedKalmanFilter(
            dim_x=n,
            dim_z=n,
            dt=dt,
            hx=hx,
            fx=fx,
            points=points,
        )
        return kf

    def do_pid(
        self, dt: float, kp: float, ki: float, kd: float, state: np.array
    ) -> float:
        """Perform PID (proportional-integral-derivative) control policy given
        a system state. This policy executes on the pendulum angle (i.e. attempts)
        to keep the pendulum in the upright position.

        Parameters
        ----------
        dt : float
            control timestep
        kp : float
            proportional gain
        ki : float
            integral gain
        kd : float
            derivative gain
        state : np.array
            input state

        Returns
        -------
        float
            control action
        """
        err = -self.wrapPi(state[2])
        errd = (err - self.prev_err) / dt
        self.integrator += err
        action = kp * err + ki * self.integrator + kd * errd
        self.prev_err = err
        return action

    def get_and_store_priors(self, state: np.ndarray, n: int) -> Tuple[int, int]:
        """utility function for getting and storing priors in some window `n`.
        This function returns two values `l` and `u` which provide the indices
        to a moving window over prior_states. You must include an attr `tick`,
        which is incremented to determine the window, as well as an attr `prior_states`
        to store states to use this function, and this function alters those attrs.

        Parameters
        ----------
        state : np.ndarray
            input state to store
        n : int
            window for returning `l`, `u` pair.

        Returns
        -------
        Tuple[int, int]
            l, u pair. If you want to get n most recent prior measurements, call
            self.prior_states[l:u]
        """
        l = max(0, self.tick - n)
        u = self.tick
        self.tick += 1
        self.prior_states.append(state)
        return l, u


class PID(Controller):
    """Basic PID Control class. This class produces a basic PID controller
    that takes the state as it is measured and executes a pid policy on it.

    Parameters
    ----------
    pid : Tuple[float, float, float]
        kp, ki, kd gains for the controller
    """

    def __init__(self, pid: Tuple[float, float, float]) -> None:
        self.kp, self.ki, self.kd = pid
        self.integrator = 0
        self.prev_err = 0

    def policy(self, state: np.ndarray, dt: float) -> Tuple[float, dict]:
        """PID policy. This just calls `do_pid` on the state, but you can
        add data collection values to it.

        Parameters
        ----------
        state : np.ndarray
            system state
        dt : float
            control timestep

        Returns
        -------
        Tuple[float, dict]
            action, data pair
        """
        action = self.do_pid(dt, self.kp, self.ki, self.kd, state)
        return action, {}


class PID_UKF(Controller):
    """PID controller with unscented kalman filter. The unscented kalman filter
    uses variance on the scale of `var_t` to get the measurement noise estimate.
    (`var_t` is a measurement of time in seconds.)


    Parameters
    ----------
    pid : Tuple[float, float, float]
        kp, ki, kd gains for the controller
    pend : [type]
        object containing the pendulum under control.
    dt : float
        control timestep
    var_t : float
        window over which to collect the measurement
    """

    def __init__(
        self, pid: Tuple[float, float, float], pend, dt: float, var_t: float
    ) -> None:
        # PID Parameters
        self.kp, self.ki, self.kd = pid
        self.integrator = 0
        self.prev_err = 0
        # UKF Parameters
        self.var_t = var_t
        self.A, self.B = self.get_linear_sys(pend.jacA, pend.jacB, dt)

        def fx(x, dt, u=0):
            return np.dot(self.A, x) + np.dot(self.B, u).T

        def hx(x):
            return x

        self.kf = self.create_ukf(dt, hx, fx)

        self.prior_states = []
        self.tick = 0

    def policy(self, state: np.ndarray, dt: float) -> Tuple[float, dict]:
        """Measure variance from prior states over `var_t`, then use that variance
        to compute an estimated state with an unscented kalman filter. The measured
        state is then used for the PID control policy.

        Parameters
        ----------
        state : np.ndarray
            system state
        dt : float
            timestep

        Returns
        -------
        Tuple[float, dict]
            action, data pair
        """
        self.get_and_store_priors(state, round(self.var_t * 1 / dt))
        prior = np.array(self.prior_states)
        if self.tick > 1:
            var = np.cov(prior.T)
        else:
            var = np.eye(4) * 1

        self.kf.Q = var ** 2
        self.kf.update(state)
        est = self.kf.x
        res = self.kf.y

        action = self.do_pid(dt, self.kp, self.ki, self.kd, state)
        self.kf.predict(dt, **{"u": action})

        labels = ["x", "xd", "t", "td"]
        data = {}
        data.update(pendsim.utils.array_to_kv("res", labels, res))
        data.update(pendsim.utils.array_to_kv("est", labels, est))
        return action, data


class LQR_UKF(Controller):
    """LQR controller with unscented kalman filter state estimation.

    Parameters
    ----------
    qr : Tuple[np.ndarray, float]
        tuple of (Q, R) matrices. In this case R is a 1x1 matrix or float.
    lqrw : int
        the window over which to perform LQR. Longer windows are more accurate,
        because they take into account a longer approximation of the system
        response. However, they take more time to compute and can be numerically
        unstable.
    pend : pendsim.sim.Pendulum
        pendulum object with jacobians of A and B matrices.
    dt : float
        control timestep
    var_t : float
        window over which to collect variances, in units of seconds.
    """

    def __init__(
        self, qr: Tuple[np.ndarray, float], lqrw: int, pend, dt: float, var_t: float
    ) -> None:
        # LQR Parameters
        self.Q, self.R = qr
        self.lqrw = lqrw
        # UKF Parameters
        self.A, self.B = self.get_linear_sys(pend.jacA, pend.jacB, dt)

        def fx(x, dt, u=0):
            return np.dot(self.A, x) + np.dot(self.B, u).T

        def hx(x):
            return x

        self.kf = self.create_ukf(dt, hx, fx)

        # Prior state info
        self.tick = 0
        self.var_t = var_t
        self.prior_states = []

    def policy(self, state: np.ndarray, dt: float) -> Tuple[float, dict]:
        self.get_and_store_priors(state, round(self.var_t * 1 / dt))
        prior = np.array(self.prior_states)
        if self.tick > 5:
            var = np.cov(prior.T)
        else:
            var = np.eye(4) * 1

        self.kf.Q = var ** 2
        self.kf.update(state)
        est = self.kf.x
        res = self.kf.y

        action = self.do_lqr(self.lqrw, self.A, self.B, self.Q, self.R, est)
        self.kf.predict(dt, **{"u": action})

        labels = ["x", "xd", "t", "td"]
        data = {}
        data.update(pendsim.utils.array_to_kv("res", labels, res))
        data.update(pendsim.utils.array_to_kv("est", labels, est))
        return action, data


class LQR(Controller):
    """Perform an LQR strategy.

    Parameters
    ----------
    pend : pendsim.sim.Pendulum
        The pendulum we want to control
    dt : float
        Timestep of the simulation
    window : int
        the window over which to perform LQR. For example; `window=5` will
        optimize over 5 timesteps
    Q : np.ndarray
        State cost array
    R : np.ndarray
        Actuation cost array
    """

    def __init__(
        self,
        pend: pendsim.sim.Pendulum,
        dt: float,
        window: int,
        Q: np.ndarray,
        R: np.ndarray,
    ) -> None:
        self.window = window
        self.A, self.B = self.get_linear_sys(pend.jacA, pend.jacB, dt)
        self.Q = Q
        self.R = R

    def policy(self, state: np.ndarray, dt: float) -> Tuple[float, dict]:
        action = self.do_lqr(self.window, self.A, self.B, self.Q, self.R, state)
        return action, {}


class LQRSwingup(Controller):
    """[summary]

    Parameters
    ----------
    pend : pendsim.sim.Pendulum
        pendulum object
    dt : float
        simulation timestep
    window : int
        the LQR window
    Q : np.ndarray
        State cost array
    R : np.ndarray
        Actuation cost array
    k : float
        swing up gain
    thresh : float, optional
        threshold for swing-up, above this angular value, the controller
        will attempt to swing itself upright; below it, it will perform
        an LQR control strategy.     by default np.pi/4
    """

    def __init__(
        self,
        pend: pendsim.sim.Pendulum,
        dt: float,
        window: int,
        Q: np.ndarray,
        R: np.ndarray,
        k: float,
        thresh: float = np.pi / 4,
    ) -> None:
        self.window = window
        self.A, self.B = self.get_linear_sys(pend.jacA, pend.jacB, dt)
        self.Q = Q
        self.R = R
        self.pend = pend
        self.thresh = thresh
        self.k = k

    def policy(self, state: np.ndarray, dt: float) -> Tuple[float, dict]:
        if self.wrapPi(state[2]) > self.thresh:
            action = self.do_swingup(self.pend, state, self.k)
        else:
            action = self.do_lqr(self.window, self.A, self.B, self.Q, self.R, state)

        return action, {}


class NoController(Controller):
    """\"No Controller\" controller. This guy does nothing, 0 action, at
    every timestep.
    """

    def __init__(self):
        pass

    def policy(self, state: np.ndarray, dt: float):
        return float(0), {}


class BangBang(Controller):
    """BangBang control strategy. If the pendulum is within a threshold given
    by `threshold`, push as hard as possible to get it to vertical. When it
    passes over `theta`, push the other way.

    This is what your air conditioner uses and it's probably not ideal for the
    fast dynamics at hand here.

    Parameters
    ----------
    magnitude : float
        magnitude of the push.
    setpoint : float, optional
        setpoint around which to push. 0 by default
    threshold : float, optional
        threshold; if the pendulum falls above the threshold (i.e. if it falls over)
        this prevents the controller from taking continous action. By default np.pi/4
    """

    def __init__(
        self, magnitude: float, setpoint: float = 0, threshold: float = np.pi / 4
    ) -> None:

        self.setpoint = setpoint
        self.magnitude = magnitude
        self.threshold = threshold

    def policy(self, state: np.ndarray, dt: float) -> Tuple[float, dict]:
        error = state[2] - self.setpoint
        action = 0.0
        if error > 0.1 and state[2] < self.threshold:
            action = -self.magnitude
        elif error < -0.1 and state[2] > -self.threshold:
            action = self.magnitude
        else:
            action = 0.0
        return action, {}
